#include "chil_global.h"

/*
 * removes pixels where velocity std is less than min_v (ground clutter)
 * and velocity is near zero.
 */
int
CHILSuppressClutter(Ray *ray, uword max_gate, float min_v)
{
  word i, j, iSV = -1, iV = -1;
  /* find which parameter is std of v, if available */
  if ((iSV = CHILFindField(ray, V_STD)) == -1) {
    return CHIL_NOTFOUND;
  }
  /* find which parameter is v, if available */
  if ((iV = CHILFindField(ray, V2)) == -1) {
    if ((iV = CHILFindField(ray, V1)) == -1) {
      return CHIL_NOTFOUND;
    }
  }

  if (max_gate > ray->ngates)
    max_gate = ray->ngates;

  /* Then remove (most) ground clutter */
  if (ray->field_type[iSV] == V_STD) {
    for (i = 0; i < max_gate; i++) {
      if (*(ray->data[iSV]+i) < min_v && *(ray->data[iSV]+i)
	  != chil_no_signal_value && *(ray->data[iV]+i) > - min_v
	  && *(ray->data[iV]+i) < min_v) {
	for (j=0;j<ray->ntotal;j++) {
	  *(ray->data[j]+i)=chil_no_signal_value;
	}
      }
    }
  }
  return CHIL_OK;
}

/*
 * Simple noise subtraction, using the far gates to get the noise level.
 * Noise is subtracted from Z only; the remaining fields are masked only.
 * If using CHILCleanRay too, call this function FIRST
 */
int
CHILSubtractNoise(Ray *ray,int noisegates,float factor)
{
  word iZ=0,i=0,j;
  double sum_noiseP=0.0,sum_noiseP2=0.0,noiseP=0.0,P,std_noiseP;
  float sfloor[MAX_FIELDS];

  if (noisegates < 1 || noisegates > ray->ngates) {
    CHILPrintf("CHILSubtractNoise: %d is an inappropriate number of noise gates\n",noisegates);
    return CHIL_ERROR;
  }

  /* find which parameter is Z */
  if ((iZ = CHILFindZH(ray)) < 0) {
    CHILPrintf("CHILSubtractNoise: attempt to subtract noise when Zh not a parameter\n");
    return CHIL_ERROR;
  }

  ray->endgate = ray->ngates - noisegates;

  /* Set the rejection thresholds for each parameter relative to that for Z */
  /* in multiplicative units */
  for (j=0;j<ray->ntotal;j++) {
    if (ray->field_type[j] == V1 || ray->field_type[j] == V2) {
      sfloor[j]=pow(10,0.1*ray->v_floor);
    }
    else if (ray->field_type[j] == V_STD) {
      sfloor[j]=pow(10,0.1*ray->sigma_v_floor);
    }
    else
      sfloor[j]=1.0;
  }


  /* estimate the noise level from the top gates */
  if (ray->range_normalised) {
    for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP = (pow(10,0.1* (*(ray->data[iZ]+i))))
	/ (ray->range[i]*ray->range[i]);
      sum_noiseP += noiseP;
      sum_noiseP2 += (noiseP * noiseP);
    }
  }
  else {
    for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP = (pow(10,0.1* (*(ray->data[iZ]+i))));
      sum_noiseP += noiseP;
      sum_noiseP2 += (noiseP * noiseP);
    }
  }

  ray->nez=10.0*log10(noiseP = sum_noiseP / noisegates);
  std_noiseP = sqrt(sum_noiseP2/noisegates - noiseP * noiseP);

  /* Subtract noise from Zh and mask other parameters*/
  if (ray->range_normalised) {
    for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      for (j=0;j<ray->ntotal;j++) {
	if (P < sfloor[j]*noiseP + factor*std_noiseP) {
	  (*(ray->data[j]+i))=chil_no_signal_value;
	}
      }
      if ((*(ray->data[iZ]+i)) != chil_no_signal_value) {
	(*(ray->data[iZ]+i)) 
	  = (float) (10.0*log10((P-noiseP)*ray->range[i]*ray->range[i]));
      }
    }
  }
  else {
    for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))));
      for (j=0;j<ray->ntotal;j++) {
	if (P < sfloor[j]*noiseP + factor*std_noiseP) {
	  (*(ray->data[j]+i))=chil_no_signal_value;
	}
      }
      if (*(ray->data[iZ]+i) != chil_no_signal_value) {
	(*(ray->data[iZ]+i))=(float) (10.0*log10(P-noiseP));
      }
    }
  }
  return CHIL_OK;
}

/*
 * More advanced noise subtraction. 
 * Assumes there is interference from some kind of oscillator and uses 
 * noisegates gates from the end of the ray to characterise it.
 * Noise is subtracted from Z only; the remaining fields are masked only.
 * If using CHILCleanRay too, call this function FIRST
 */
int
CHILSubtractOscillatorNoise(Ray *ray,int noisegates,float factor)
{
  word iZ=0,i=0,j;
  double sum_diff_noiseP2=0.0,noiseP[1023],P,std_noiseP,sum_noiseP=0;
  float sfloor[MAX_FIELDS];

  if (noisegates <1 || noisegates > ray->ngates) {
    CHILPrintf("CHILSubtractNoise: %d is an inappropriate number of noise gates\n",noisegates);
    return CHIL_ERROR;
  }

  /* find which parameter is Z */
  while (ray->field_type[iZ] != ZH && iZ<ray->ntotal)
    iZ++;
  if (ray->field_type[iZ] != ZH) {
    CHILPrintf("CHILSubtractNoise: attempt to subtract noise when Zh not a parameter\n");
    return CHIL_ERROR;
  }

  ray->endgate = ray->ngates - noisegates;

  /* Set the rejection thresholds for each parameter relative to that for Z */
  /* in multiplicative units */
  for (j=0;j<ray->ntotal;j++) {
    if (ray->field_type[j] == V1 || ray->field_type[j] == V2) {
      sfloor[j]=pow(10,0.1*ray->v_floor);
    }
    else if (ray->field_type[j] == V_STD) {
      sfloor[j]=pow(10,0.1*ray->sigma_v_floor);
    }
    else
      sfloor[j]=1.0;
  }

  /* get the noise level from the top gates */
  if (ray->range_normalised) {
    for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP[i] = (pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      sum_noiseP += noiseP[i];
    }
  }
  else {
    for (i=ray->ngates-noisegates;i<ray->ngates;i++) {
      noiseP[i] = (pow(10,0.1* (*(ray->data[iZ]+i))));
      sum_noiseP += noiseP[i];
    }
  }

  ray->nez=10.*log10(sum_noiseP/noisegates);

  /* estimate the standard deviation from the difference between adjacent gates */
  for (i=ray->ngates-noisegates+1;i<ray->ngates;i++) {
    sum_diff_noiseP2 += (noiseP[i]-noiseP[i-1])*(noiseP[i]-noiseP[i-1]);
  }
  std_noiseP=sqrt(sum_diff_noiseP2/(noisegates-1));

  /* replicate the noise pattern down through the other gates */
  for (i=ray->ngates-noisegates-1;i>=0;i--) {
    noiseP[i]=noiseP[i+noisegates];
  }

  /* Subtract noise from Zh and mask other parameters*/
  if (ray->range_normalised) {
    for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))))/(ray->range[i]*ray->range[i]);
      for (j=0;j<ray->ntotal;j++) {
	if (P < sfloor[j]*noiseP[i] + factor*std_noiseP) {
	  (*(ray->data[j]+i))=chil_no_signal_value;
	}
      }
      if (*(ray->data[iZ]+i) != chil_no_signal_value) {
	(*(ray->data[iZ]+i))=(float) (10.0*log10((P-noiseP[i])*ray->range[i]*ray->range[i]));
      }
    }
  }
  else {
    for (i=0;i<ray->ngates;i++) {
      P=(pow(10,0.1* (*(ray->data[iZ]+i))));
      for (j=0;j<ray->ntotal;j++) {
	if (P < sfloor[j]*noiseP[i] + factor*std_noiseP) {
	  (*(ray->data[j]+i))=chil_no_signal_value;
	}
      }
      if (*(ray->data[iZ]+i) != chil_no_signal_value) {
	(*(ray->data[iZ]+i))=(float) (10.0*log10(P-noiseP[i]));
      }
    }
  }
  return CHIL_OK;
}

/* Checks to see if the specified field (e.g. ZH) exceeds maxval at 
   the specified gate of the ray. If so, CHIL_ERROR is returned,
   otherwise CHIL_OK is returned. If the gate or field is out of 
   range, CHIL_NOTFOUND is returned. WARNING: be careful to use
   this function at the appropriate stage in the processing... */
int
CHILIsCorrupt(Ray *ray, uword gate, uchar field, float maxval)
{
  int ifield, val = 0;
  if (gate >= ray->ntotal || (ifield = CHILFindField(ray, field)) < 0) {
    return CHIL_NOTFOUND;
  }

  if ((val = *(ray->data[ifield]+gate)) > maxval 
      && val != chil_no_signal_value) {
    return CHIL_ERROR;
  }
  return CHIL_OK;
}

/* Looks for gate that has a Z more than maxjump greater than that two gates
   below; if so returns CHIL_ERROR, else returns CHIL_OK. If no Z, returns
   CHIL_NOTFOUND */
int
CHILIsRangeGlitched(Ray *ray, uword maxgate, float maxjump)
{
  int ifield = CHILFindZH(ray);
  uword i;
  if (ifield < 0) {
    return CHIL_NOTFOUND;
  }
  for (i = maxgate; i < ray->ngates - 2; i++) {
    if (*(ray->data[ifield]+i) + maxjump < *(ray->data[ifield]+i+2)) {
      return CHIL_ERROR;
    }
  }
  return CHIL_OK;
}

/* Checks that the clutter (or transmit pulse) occupies more than
   min_clutter_gates.  Assumes that noise has already been subtracted.
   Returns CHIL_ERROR if low sensitivity, CHIL_OK for good data,
   CHIL_NOTFOUND if no Z in dataset. */
int
CHILIsLowSensitivity(Ray *ray, uword min_clutter_gates)
{
  int ifield = CHILFindZH(ray);
  uword i = 0;
  if (ifield < 0) {
    return CHIL_NOTFOUND;
  }
  /* Find first proper gate */
  while (ray->data[ifield][i] == chil_no_signal_value && i < ray->ngates -1)
    i++;

  if (ray->data[ifield][i] != chil_no_signal_value) {
    uword j = i+1;
    while (ray->data[ifield][j] != chil_no_signal_value && j < ray->ngates -1)
      j++;
    if (j-i >= min_clutter_gates) {
      return CHIL_OK;
    }
    else if (j < ray->ngates - 3
	     && (ray->data[ifield][j+1] != chil_no_signal_value
		 || ray->data[ifield][j+2] != chil_no_signal_value)) {
      return CHIL_OK;
    }
  }

  return CHIL_ERROR;
}


#define MAX_GATES 1024
/* Looks for a glitched *unaveraged* ray, caused by some bug in the PCI system.
   This is a complete kludge. A real mess. */
int
CHILIsPCIGlitched(RasterInfo *info, Ray *ray)
{
  int zh_index = CHILFindZH(ray);
  float (*zh_func) (short) = info->field_function[zh_index];
  int gate;
  static int first_gate = 6, last_gate = 17;
  float zh[MAX_GATES];
  static char warned = 0;

  if (last_gate + 1 > MAX_GATES) {
    if (!warned) {
      CHILPrintf("Warning: too many gates to do PCI clean\n");
      warned = 1;
    }
    return 0;
  }

  /* Get some Zh values to work with */
  zh[0] = zh_func(ray->buffer[zh_index]);
  for (gate = first_gate; gate <= last_gate; ++gate) {
    if ((zh[gate] = zh_func(ray->buffer[gate*(ray->nrawfields)+zh_index]))
	< MIN_ZH) {
      /*      zh[gate] = zh[0]; // This doesn't work at all!!! */
      zh[gate] = -99.0;
    }
  }

  /*
  if (zh[7] < -55.0) {
    return 0;
  }
  else
  */
  if (zh[7] + zh[8] + zh[10] + zh[11] + zh[13] + zh[14]
      - 1.5 * (zh[6] + zh[9] + zh[12] + zh[15]) > 20) {
    return 1;
  }
  else {
    return 0;
  }
}
