#include "chil_global.h"

float chil_no_signal_value = -999.0;

void
CHILSetNoSignalValue(float no_signal_value)
{
  chil_no_signal_value = no_signal_value;
}

/* Return the index of a particular field (ZH, ZDR etc.) in a ray,
 * or -1 if not present */
int
CHILFindField(Ray *ray, int field)
{
  int i;
  for (i=0;i<ray->ntotal;i++) {
    if (ray->field_type[i] == field) {
      return i;
    }
  }
  return -1;
}


/* Simple calibration of the Zh field (and nez value) */
int
CHILCalibrateZh(Ray *ray,float Zcal)
{
  word iZ=0,i=0;

  if ((iZ = CHILFindZH(ray)) < 0) {
#ifdef DEBUG
    CHILPrintf("CHILCalibrateZh: attempt to calibrate Zh when Zh not a parameter\n");
#endif
    return CHIL_NOTFOUND;
  }
  for (i=0;i<ray->ngates;i++) {
    if (*(ray->data[iZ]+i) != chil_no_signal_value) 
      (*(ray->data[iZ]+i)) += Zcal;
  }
  ray->nez += Zcal;
  return CHIL_OK;
}

/* Simple calibration of an arbitrary field (but not nez value) */
int
CHILCalibrateField(Ray *ray, int fieldid, float cal)
{
  word ip=0,i=0;

  if ((ip = CHILFindField(ray, fieldid)) < 0) {
#ifdef DEBUG
    CHILPrintf("CHILCalibrateField: field with id %d not found in ray\n", fieldid);
#endif
    return CHIL_NOTFOUND;
  }
  for (i=0;i<ray->ngates;i++) {
    if (*(ray->data[ip]+i) != chil_no_signal_value) 
      (*(ray->data[ip]+i)) += cal;
  }
  return CHIL_OK;
}


/* 
 * Simple cleanup, to be applied AFTER CHILSubtractNoise (if used).
 * Removes `isolated' pixels in ray which gets rid of some of those 
 * annoying lines in cloud data. Uses Z, or the first field in the
 * data.
 */
int
CHILCleanRay(Ray *ray)
{
  int ifield = 0;
  word i,j;

  if ((ifield = CHILFindZH(ray)) < 0) {
    ifield = 0;
  }

  /* Remove isolated pixels */
  for (i=1;i<ray->ngates-1;i++) {
    if (*(ray->data[ifield]+i-1) == chil_no_signal_value 
	&& *(ray->data[ifield]+i+1) == chil_no_signal_value) {
      for (j=0;j<ray->ntotal;j++) {
	*(ray->data[j]+i)=chil_no_signal_value;
      }
    }
  }

  /* Remove isolated pixels in LDR */
  if ((ifield = CHILFindField(ray, LDR)) > 0) {
    for (i=1;i<ray->ngates-1;i++) {
      if (*(ray->data[ifield]+i-1) == chil_no_signal_value 
	  && *(ray->data[ifield]+i+1) == chil_no_signal_value) {
	*(ray->data[ifield]+i)=chil_no_signal_value;
      }
    }
  }

  return CHIL_OK;
}

/* 
 * Same as CHILCleanRay() but removes both isolated single pixels 
 * and isolated pairs of pixels.
 */
int
CHILDoubleCleanRay(Ray *ray)
{
  int ifield = 0;
  word i,j;

  if ((ifield = CHILFindZH(ray)) < 0) {
    ifield = 0;
  }

  /* Remove isolated pairs of pixels */
  for (i=1;i<ray->ngates-2;i++) {
    if (*(ray->data[ifield]+i-1) == chil_no_signal_value 
	&& (*(ray->data[ifield]+i+1) == chil_no_signal_value
	    || *(ray->data[ifield]+i+2) == chil_no_signal_value)) {
      for (j=0;j<ray->ntotal;j++) {
	*(ray->data[j]+i)=chil_no_signal_value;
      }
    }
  }

  /* Remove isolated pairs of pixels in LDR */
  if ((ifield = CHILFindField(ray, LDR)) > 0) {
    for (i=1;i<ray->ngates-2;i++) {
      if (*(ray->data[ifield]+i-1) == chil_no_signal_value 
	  && (*(ray->data[ifield]+i+1) == chil_no_signal_value
	      || *(ray->data[ifield]+i+2) == chil_no_signal_value)) {
	*(ray->data[ifield]+i)=chil_no_signal_value;
      }
    }
  }

  return CHIL_OK;
}

/*
 * Removes up to <cleangates> isolated gates every <cycle> gates, at
 * <offset> from the first gate. This is for removing dodgy lines
 * caused by an interference effect.
 */
int
CHILCustomCleanRay(Ray *ray, uint cycle, uint offset, uint cleangates)
{
  int ifield = 0;
  word i, j, k;

  if ((ifield = CHILFindZH(ray)) < 0) {
    ifield = 0;
  }

  /* Remove isolated pixels */
  for (i = offset+1; i < ray->ngates-cleangates-1; i += cycle) {
    if (*(ray->data[ifield]+i-1) == chil_no_signal_value
	&& *(ray->data[ifield]+i+cleangates) == chil_no_signal_value) {
      for (j = 0; j < ray->ntotal; ++j) {
	for (k = i; k < i+cleangates; ++k) {
	  *(ray->data[j]+k) = chil_no_signal_value;
	}
      }
    }
  }
  return CHIL_OK;
}

/*
 * Interpolates over <cleangates> isolated gates every <cycle> gates, at
 * <offset> from the first gate. This is for removing dodgy lines
 * caused by an interference effect.
 */
int
CHILInterpolateRay(Ray *ray, uint cycle, uint offset, uint cleangates)
{
  int ifield;
  word i, k;

  for (i = offset+1; i < ray->ngates-cleangates-1; i += cycle) {
    for (ifield = 0; ifield < ray->ntotal; ++ifield) {
      float val1 = *(ray->data[ifield]+i-1);
      float val2 = *(ray->data[ifield]+i+cleangates);
      if (val1 == chil_no_signal_value || val2 == chil_no_signal_value) {
	for (k = i; k < i+cleangates; ++k) {
	  *(ray->data[ifield]+k) = chil_no_signal_value;
	}
      }
      else if (ray->field_type[ifield] == V1 && ray->use_v1fold) {
	double x1 = cos(M_PI * val1 / ray->v1fold);
	double y1 = sin(M_PI * val1 / ray->v1fold);
	double x2 = cos(M_PI * val2 / ray->v1fold);
	double y2 = sin(M_PI * val2 / ray->v1fold);
	double x, y;
	int l = 1;
	for (k = i; k < i+cleangates; ++k, ++l) {
	  x = (x1 * (cleangates+1-l) + x2 * l) / (cleangates+1);
	  y = (y1 * (cleangates+1-l) + y2 * l) / (cleangates+1);
	  *(ray->data[ifield]+k) = atan2(y,x) * ray->v1fold * M_1_PI;
	}
      }
      else if (ray->field_type[ifield] == V2 && ray->use_v2fold) {
	double x1 = cos(M_PI * val1 / ray->v2fold);
	double y1 = sin(M_PI * val1 / ray->v2fold);
	double x2 = cos(M_PI * val2 / ray->v2fold);
	double y2 = sin(M_PI * val2 / ray->v2fold);
	double x, y;
	int l = 1;
	for (k = i; k < i+cleangates; ++k, ++l) {
	  x = (x1 * (cleangates+1-l) + x2 * l) / (cleangates+1);
	  y = (y1 * (cleangates+1-l) + y2 * l) / (cleangates+1);
	  *(ray->data[ifield]+k) = atan2(y,x) * ray->v2fold * M_1_PI;
	}
      }
      else {
	int l = 1;
	for (k = i; k < i+cleangates; ++k, ++l) {
	  *(ray->data[ifield]+k)
	    = (val1 * (cleangates+1-l) + val2 * l) / (cleangates+1);
	}
      }
    }
  }
  return CHIL_OK;
}


/*
 * The following function performs range correction;
 * This involves correcting Z according to what's in the range_offset
 * member of RasterInfo 
*/
int
CHILRangeCorrection(RasterInfo *info,Ray *ray)
{
  word iZ=0,i;
  float range_offset;

  range_offset=0.001*ray->range_offset;

  if (ray->correct_range) {
    CHILPrintf("Range normalisation already performed\n");
    return CHIL_NOTDONE;
  }

  if (range_offset==0) {
    CHILPrintf("Attempt to range normalise when range offset is unknown\n");
    return CHIL_NOTDONE;
  }

  /* find which parameter is Z */
  iZ = CHILFindZH(ray);

  if (iZ >= 0) {
    if (ray->range_normalised) {
      for (i=0;i<ray->ngates;i++) {
	if (ray->range[i]+range_offset > 0.0 && 
	    *(ray->data[iZ]+i) != chil_no_signal_value) {
	  *(ray->data[iZ]+i) +=
	    20.0*log10((ray->range[i]+range_offset)/(ray->range[i]));
	}
	else {
	  *(ray->data[iZ]+i)=chil_no_signal_value;
	}
      }
    }
    else {
      for (i=0;i<ray->ngates;i++) {
	if (ray->range[i]+range_offset > 0.0 && 
	    *(ray->data[iZ]+i) != chil_no_signal_value) {
	  *(ray->data[iZ]+i) +=
	    20.0*log10(ray->range[i]+range_offset);
	}
	else
	  *(ray->data[iZ]+i)=chil_no_signal_value;
      }
    }
  }
  /* Set ray->startgate to the index of the first non-negative range
     gate.  We actually use 0.5 metres as the minimum as round-off
     error can result in the first gate being small and negative. */
  for (i=0;i<ray->ngates;i++) {
    if (ray->range[i]+range_offset >= 0.0005) {
      ray->startgate = i;
      break;
    }
  }
  ray->correct_range=1;
  return CHIL_OK;
}

/* Negate all velocities; i.e. use the opposite sign convention for velocity */
int
CHILReverseVelocities(Ray *ray)
{
  word i, iv;
  if ((iv = CHILFindField(ray, V1)) > -1) {
    float *v = ray->data[iv];
    for (i = 0; i < ray->ngates; ++i, ++v) {
      if (*v != chil_no_signal_value) {
	*v = -(*v);
      }
    }
  }
  if ((iv = CHILFindField(ray, V2)) > -1) {
    float *v = ray->data[iv];
    for (i = 0; i < ray->ngates; ++i, ++v) {
      if (*v != chil_no_signal_value) {
	*v = -(*v);
      }
    }
  }
  return 0;
}
